/******************************************************************************/
/******************************************************************************/
/* Program: xa_mqs_oracle                                                     */
/* Author:  Predrag Maksimovic - Max                                          */
/* Date:    10/21/2001                                                        */
/* Contact: pmaksimovic@2d3d.com                                              */
/******************************************************************************/
/* This example shows how to use MQSeries and Oracle in a distributed         */
/* transaction. MQSeries queue manager serves as a TX transaction manager     */
/* and as an XA resource manager, while Oracle DBMS serves as an XA resource  */
/* manager only                                                               */
/* MQI API is used to access MQSeries, and OCI API is used to access Oracle   */
/*                                                                            */
/* The program arguments are:                                                 */
/* #1 Name of the queue manager (the TX transaction manager)                  */
/* #2 Name of the queue from which a message is received                      */
/* #3 Name of the table to which the message is inserted                      */
/* #4 Name of the actual column to which the message is inserted              */
/*                                                                            */
/* The program receives a message from the specified queue, and inserts it    */
/* into the specified database table - all within the same distributed        */
/* transaction                                                                */
/*----------------------------------------------------------------------------*/
/* Change the following function calls:                                       */
/*                                                                            */
/*    envhp = xaoEnv("dbname");                                               */
/*    svchp = xaoSvcCtx("dbname");                                            */
/*                                                                            */
/* where dbname is the same value specified for +DB=... in the XAOpenString   */
/*                                                                            */
/******************************************************************************/
/* Resource manager stanza:                                                   */
/* XAResourceManager:                                                         */
/* Name=oracle_rm_name                                                        */
/* SwitchFile=switch_load_file_full_path                                      */
/* XAOpenString=Oracle_XA+Acc=P/usr/pswd+SesTm=35+SqlNet=connectionstring     */
/* XACloseString=This_is_Close_String                                         */
/* ThreadOfControl=PROCESS                                                    */
/******************************************************************************/
/* Hint1: Replace oracle_rm_name, switch_load_file_full_path, usr, pswd and   */
/*        connectionstring with the actual values                             */
/*        oracle_rm_name: user defined name for this RM (any name is fine)    */
/*        switch_load_file_full_path: path to the switch load library         */
/*        usr: Oracle username                                                */
/*        pswd: Oracle password                                               */
/*        connectionstring: Oracle connection string                          */
/* Hint2: Read MQSeries documentation on Oracle configuration:                */
/*        Application Programming Guide                                       */
/*        System Administration                                               */
/*        These books can be downloaded from the IBM's site, and look for     */
/*        the chapters on transactions                                        */
/* Hint2: This close string is required in order to avoid the annoying        */
/*        xa_close error messages in the queue manager error log files        */
/* Hint3: On WinNT, link this example program with mqm.lib, oci.lib and       */
/*        oraxa8.lib (Oracle8.x)                                              */
/*                                                                            */
/******************************************************************************/
/******************************************************************************/
/******************************************************************************/
/* Include general header files                                               */
/******************************************************************************/
#include <stdio.h>
#include <string.h>
/******************************************************************************/
/* Include MQSeries header files                                              */
/******************************************************************************/
#include <cmqc.h>
/******************************************************************************/
/* Include Oracle header files                                                */
/******************************************************************************/
#include <oci.h>
#include <ocidem.h>

/******************************************************************************/
/* Function: DisplayMQSeriesError                                             */
/* Displays MQSeries MQI error info                                           */
/******************************************************************************/
void DisplayMQSeriesError(MQLONG lReason)
{
    printf("MQI call failed with reason code %d\n", lReason);
}

/******************************************************************************/
/* Function: DisplayOracleError                                               */
/* Displays Oracle OCI error info                                             */
/******************************************************************************/
void DisplayOracleError(OCIError* errhp, sword status)
{
    text errbuf[512];
    sb4 errcode = 0;

    switch (status)
    {
        case OCI_SUCCESS_WITH_INFO:
            printf("The error is: OCI_SUCCESS_WITH_INFO\n");
            break;
        case OCI_INVALID_HANDLE:
            printf("The error is: OCI_INVALID_HANDLE\n");
            break;
        case OCI_STILL_EXECUTING:
            (void) printf("The error is: OCI_STILL_EXECUTE\n");
            break;
        case OCI_CONTINUE:
            (void) printf("The error is: OCI_CONTINUE\n");
            break;
        case OCI_ERROR:
            (void) OCIErrorGet((dvoid*)errhp, (ub4)1, (text*)NULL, &errcode,
                                errbuf, (ub4)sizeof(errbuf), OCI_HTYPE_ERROR);
            (void) printf("The error is: %.*s\n", 512, errbuf);
            break;
        default:
            printf("Error details could not be obtained\n");
            break;
    }

}

/******************************************************************************/
/* Function: main                                                             */
/* Program entry point                                                        */
/******************************************************************************/
int main(int argc, char *argv[])                 
{
   /***************************************************************************/
   /* MQSeries variables                                                      */
   /***************************************************************************/
   MQOD     od              = {MQOD_DEFAULT};    /* object descriptor         */
   MQMD     md              = {MQMD_DEFAULT};    /* message descriptor        */
   MQGMO    gmo             = {MQGMO_DEFAULT};   /* get message options       */
   MQBO     bo              = {MQBO_DEFAULT};    /* begin options             */
   MQHCONN  hConn           = 0;                 /* handle to connection      */
   MQHOBJ   hObj            = 0;                 /* handle to object          */
   MQLONG   lOptions        = 0;                 /* options                   */
   MQLONG   lCompCode       = 0;                 /* completion code           */
   MQLONG   lReason         = 0;                 /* reason code               */
   MQLONG   lMsgBufLen      = 100;               /* message buffer length     */
   MQLONG   lMsgLen         = 0;                 /* message length received   */
   MQBYTE   byMsgBuf[100];                       /* message buffer (100 bytes)*/
   char*    szQueueManager  = NULL;              /* queue manager name        */
   char*    szQueue         = NULL;              /* queue name                */

   /***************************************************************************/
   /* Oracle variables                                                        */
   /***************************************************************************/
   OCIEnv*       envhp              = NULL;       /* environm. handle pointer */
   OCISvcCtx*    svchp              = NULL;       /* service handle pointer   */
   OCIStmt*      stmthp             = NULL;       /* statement handle pointer */
   OCIError*     errhp              = NULL;       /* error handle pointer     */
   sword         status             = OCI_SUCCESS;/* return code              */
   char*         szConnectionString = NULL;       /* Oracle connection string */
   char*         szTable            = NULL;       /* table name               */
   char*         szColumn           = NULL;       /* column name              */
   text          szSQLStatement[200];             /* SQL statement            */

   /***************************************************************************/
   /* general variables                                                       */
   /***************************************************************************/
   char     cSelection      = 'c';               /* end transaction selection */

   /***************************************************************************/
   /* Check input arguments                                                   */
   /***************************************************************************/
   if (argc != 5)
   {
       printf("Usage: %s queue_manager_name queue_name "
              "table_name column_name\n",
              argv[0]);
       goto Exit_Program;
   }
   else
   {
       szQueueManager     = argv[1];
       szQueue            = argv[2];
       szTable            = argv[3];
       szColumn           = argv[4];
   }

   
   /***************************************************************************/
   /* Connect to queue manager                                                */
   /***************************************************************************/
   MQCONN(szQueueManager,
          &hConn,
          &lCompCode,
          &lReason);

   if (lCompCode != MQCC_OK)
   {
       printf("MQCONN() failed\n");
       DisplayMQSeriesError(lReason);
       goto Exit_Program;
   }
   else
   {
       printf("Connected to queue manager %s\n", szQueueManager);
   }

    
   /***************************************************************************/
   /* Start a new transaction                                                 */
   /***************************************************************************/
   MQBEGIN (hConn, &bo, &lCompCode, &lReason);

   if (lCompCode != MQCC_OK)
   {
       printf("MQBEGIN() failed\n");
       DisplayMQSeriesError(lReason);
       goto Exit_Program;
   }
   else
   {
       printf("Unit of work started\n");
   } 

   /***************************************************************************/
   /* Open the queue                                                          */
   /***************************************************************************/
   strncpy(od.ObjectName, szQueue, MQ_Q_NAME_LENGTH);
   lOptions = MQOO_INPUT_AS_Q_DEF + MQOO_FAIL_IF_QUIESCING;
   MQOPEN(hConn, &od, lOptions, &hObj, &lCompCode, &lReason);
   if (lCompCode != MQCC_OK)
   {
       printf("MQOPEN() failed\n");
       DisplayMQSeriesError(lReason);
       goto Exit_Program;
   }
   else
   {
       printf("Opened queue %s\n", szQueue);
   } 


   /***************************************************************************/
   /* Initialize Oracle (OCI) - this is not required here                     */
   /***************************************************************************/

   printf("Initializing OCI library\n");
   status = OCIInitialize((ub4)OCI_DEFAULT, (dvoid *)0,
                          (dvoid* (*)(dvoid*, size_t))0,
                          (dvoid* (*)(dvoid*, dvoid*, size_t))0,
                          (void (*)(dvoid*, dvoid*))0);
   if (status != OCI_SUCCESS)
   {
       printf("OCIInitialize() failed\n");
       goto Exit_Program;
   }

   /***************************************************************************/
   /* Obtain the environment handle for the XA connection                     */
   /***************************************************************************/
   envhp = xaoEnv("dbname");
   if (envhp == NULL)
   {
       printf("xaoEnv() failed");
       goto Exit_Program;
   }
   else
   {
       printf("Environment handle(XA) obtained\n");
   }
   /***************************************************************************/
   /* Obtain the service context handle for the XA connection                 */
   /***************************************************************************/
   svchp = xaoSvcCtx("dbname");
   if (svchp == NULL)
   {
       printf("xaoSvcCtx() failed");
       goto Exit_Program;
   }
   else
   {
       printf("Service context handle(XA) obtained\n");
   }

   /***************************************************************************/
   /* Allocate an error report handle                                         */
   /***************************************************************************/
   status = OCIHandleAlloc((dvoid*)envhp, (dvoid**)&errhp, OCI_HTYPE_ERROR, 
                           (size_t)0, (dvoid**)0);
   if (status != OCI_SUCCESS)
   {
       printf("Could not allocate error report handle\n");
       goto Exit_Program;
   }
   else
   {
       printf("Error report handle successfully allocated\n");
   }

   /***************************************************************************/
   /* Allocate a SQL statement handle                                         */
   /***************************************************************************/
   status = OCIHandleAlloc((dvoid*)envhp, (dvoid**)&stmthp, OCI_HTYPE_STMT,
                           (size_t)0, (dvoid **)0);
   if (status != OCI_SUCCESS)
   {
       printf("Could not allocate SQL statement handle\n");
       goto Exit_Program;
   }
   else
   {
       printf("SQL statement handle successfully allocated\n");
   }

   /***************************************************************************/
   /* Get a message                                                           */
   /***************************************************************************/
   gmo.Options = MQGMO_WAIT + MQGMO_SYNCPOINT;
   memset(byMsgBuf, 0, lMsgBufLen);
   MQGET(hConn, hObj, &md, &gmo,
         lMsgBufLen, byMsgBuf, &lMsgBufLen,
         &lCompCode, &lReason);
   if (lCompCode != MQCC_OK)
   {
       printf("MQGET failed\n");
       DisplayMQSeriesError(lReason);
       goto Exit_Program;
   }
   else
   {
       printf("Message retrieved from the queue %s\n", szQueue);
   }

   /***************************************************************************/
   /* Build the SQL statement                                                 */
   /***************************************************************************/
   sprintf(szSQLStatement, "INSERT INTO %s(%s) VALUES('%s')",
           szTable, szColumn, (char*)byMsgBuf);
   printf("SQL statement to be executed: %s\n", szSQLStatement);

   
   /***************************************************************************/
   /* Prepare the statement                                                   */
   /***************************************************************************/
   status = OCIStmtPrepare(stmthp, errhp, szSQLStatement,
                           (ub4)strlen((char*)szSQLStatement),
                           (ub4)OCI_NTV_SYNTAX, (ub4)OCI_DEFAULT);
   if (status == OCI_SUCCESS)
   {
       printf("SQL statement successfully prepared\n");
   }
   else
   {
       DisplayOracleError(errhp, status);
       goto Exit_Program;
   }
   
   /***************************************************************************/
   /* Execute the statement                                                   */
   /***************************************************************************/
   status = OCIStmtExecute(svchp, stmthp, errhp, (ub4)1, (ub4)0,
               (CONST OCISnapshot*)NULL, (OCISnapshot*)NULL, OCI_DEFAULT);
   if (status == OCI_SUCCESS)
   {
       printf("SQL statement successfully executed\n");
   }
   else
   {
       DisplayOracleError(errhp, status);
       goto Exit_Program;
   }

   /***************************************************************************/
   /* End the transaction                                                     */
   /***************************************************************************/
   printf("Enter C or c to commit, any other key to rollback\n");
   cSelection = getchar();
   if (cSelection == 'c' || cSelection == 'C')
   {
       /***********************************************************************/
       /* Commit the transaction                                              */
       /***********************************************************************/
       MQCMIT(hConn, &lCompCode, &lReason);
       if (lCompCode != MQCC_OK)
       {
           printf("MQCMIT() failed\n");
           DisplayMQSeriesError(lReason);
           goto Exit_Program;
       }
       else
       {
           printf("The transaction has been committed\n");
       }
   }
   else
   {
       /***********************************************************************/
       /* Rollback the transaction                                            */
       /***********************************************************************/
       MQBACK(hConn, &lCompCode, &lReason);
       if (lCompCode != MQCC_OK)
       {
           printf("MQBACK() failed\n");
           DisplayMQSeriesError(lReason);
           goto Exit_Program;
       }
       else
       {
           printf("The transaction has been rolled back\n");
       }
   }



Exit_Program:

   /***************************************************************************/
   /* Try to free the statement handle                                        */
   /***************************************************************************/
   if (stmthp)
   {
       status = OCIHandleFree(stmthp, OCI_HTYPE_STMT);
       if (status != OCI_SUCCESS)
       {
           printf("OCIHandleFree(OCI_HTYPE_STMT) failed\n");
       }
       else
       {
           printf("Freed SQL statement handle\n");
       }
   }

   /***************************************************************************/
   /* Try to free the error report handle                                     */
   /***************************************************************************/
   if (errhp)
   {
       status = OCIHandleFree(errhp, OCI_HTYPE_ERROR);
       if (status != OCI_SUCCESS)
       {
           printf("OCIHandleFree(OCI_HTYPE_ERROR) failed\n");
       }
       else
       {
           printf("Freed error report handle\n");
       }
   }

   /***************************************************************************/
   /* Service context handle and environment handle obtained for the XA       */
   /* connection need not to be freed                                         */
   /* These are special handles and MQDISC() is taking care of them (by       */
   /* communicating with the Oracle client library                            */
   /* Freeing these handles explicitly may cause core on some platforms       */
   /***************************************************************************/
   /****************************************************************************
   if (svchp)
   {
       status = OCIHandleFree(svchp, OCI_HTYPE_SVCCTX);
       if (status != OCI_SUCCESS)
       {
           printf("OCIHandleFree(OCI_HTYPE_SVCCTX) failed\n");
       }
       else
       {
           printf("Freed service context handle\n");
       }
   }

   /***************************************************************************/
   /* Try to free the environment handle                                      */
   /***************************************************************************/
   /*
   if (envhp)
   {
       status = OCIHandleFree(envhp, OCI_HTYPE_ENV);
       if (status != OCI_SUCCESS)
       {
           printf("OCIHandleFree(OCI_HTYPE_ENV) failed\n");
       }
       else
       {
           printf("Freed environment handle\n");
       }
   }
   ****************************************************************************/
   /***************************************************************************/
   /* Close queue if opened                                                   */
   /***************************************************************************/
   if (hObj)
   {
      lOptions = 0;
      MQCLOSE(hConn, &hObj, lOptions, &lCompCode, &lReason);
     
      if (lCompCode != MQCC_OK)
      {
          printf("MQCLOSE() failed\n");
          DisplayMQSeriesError(lReason);
      }
      else
      {
          printf("Closed queue %s\n", szQueue);
      }
   }

   /***************************************************************************/
   /* Disconnect from queue manager if connected                              */
   /***************************************************************************/
   if (hConn)
   {
      MQDISC(&hConn, &lCompCode, &lReason);

      if (lCompCode != MQCC_OK)
      {
          printf("MQDISC() failed\n");
          DisplayMQSeriesError(lReason);
      }
      else
      {
          printf("Disconnected from queue manager %s\n", szQueueManager);
      }
   }

   /***************************************************************************/
   /* Exit program                                                            */
   /***************************************************************************/
   return 0;
}
